import pysisu
import os

API_KEY = os.environ.get('SISU_API_KEY')
ANALYSIS_ID = 128998

table = pysisu.get_results(ANALYSIS_ID, API_KEY, {"top_drivers": "False"})
print(','.join([x.column_name for x in table.header]))
for row in table.rows:
    print(row)



#
# Copyright 2022 Sisu Data, Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Type
import requests
import urllib
from pysisu.sisu.v1.api import Factor, FactorValue, KeyDriverAnalysisResultGroupComparison, KeyDriverAnalysisResultSubgroup, KeyDriverAnalysisResultTimeComparison, LatestAnalysisResultResponse
import datetime


class safelist(list):
    # taken from https://stackoverflow.com/questions/5125619/why-doesnt-list-have-safe-get-method-like-dictionary
    def get(self, index, default=None):
        try:
            return self.__getitem__(index)
        except IndexError:
            return default


@dataclass
class Table:
    header: List["HeaderColumn"]
    rows: List["Row"]


@dataclass
class HeaderColumn:
    column_name: str
    column_type: Type


@dataclass
class Row:
    subgroup_id: int
    is_top_driver: bool
    factor_0_dimension: str
    factor_0_value: str
    factor_1_dimension: str
    factor_1_value: str
    factor_2_dimension: str
    factor_2_value: str

    def __str__(self):
        variables = []
        for x in vars(self).values():
            if isinstance(x, (int, float)):
                variables.append(x)
            else:
                variables.append(f"'{x}'")
        return ','.join([str(x) for x in variables])


@dataclass
class TimeCompareRow(Row):
    previous_period_size: float
    recent_period_size: float
    previous_period_value: float
    recent_period_value: float
    previous_period_start: datetime.datetime
    previous_period_end: datetime.datetime
    recent_period_start: datetime.datetime
    recent_period_end: datetime.datetime


@dataclass
class GroupCompareRow(Row):
    group_a_size: float
    group_b_size: float
    group_a_value: float
    group_b_value: float
    group_a_name: str
    group_b_name: str


@dataclass
class GeneralPerformanceRow(Row):
    size: float
    value: float


def build_url(base_url, path, args_dict):
    url_parts = list(urllib.parse.urlparse(base_url))
    url_parts[2] = path
    url_parts[4] = urllib.parse.urlencode(args_dict)
    return urllib.parse.urlunparse(url_parts)


def get_factor_value(factor: Factor) -> str:
    value: FactorValue = factor.value
    val = (
        value.boolean_value or
        value.float_value or
        value.integer_value or
        value.string_value or
        value.timestamp_value
    )
    return str(val)


@dataclass
class FactorDimVal:
    dimension: Optional[str]
    value: Optional[any]


def get_factor(dimension: str, factor: Factor) -> FactorDimVal:
    if factor.bin:
        return FactorDimVal(dimension, f'{factor.bin.lower_bound_percentile} to {factor.bin.upper_bound_percentile}')
    elif factor.value is not None:
        return FactorDimVal(dimension, get_factor_value(factor))
    elif factor.keyword:
        return FactorDimVal(dimension, factor.keyword.keyword)
    else:
        raise ValueError("invalid factor")


def get_factors(factors: Dict[str, Factor]) -> Tuple[FactorDimVal, FactorDimVal, FactorDimVal]:
    converted_factors = safelist([get_factor(dim, fact)
                                  for dim, fact in factors.items()])

    empty_factor = FactorDimVal(None, None)
    return converted_factors.get(0), converted_factors.get(1, empty_factor), converted_factors.get(2, empty_factor)


def build_header_from_row(rows: List[Row]) -> List[HeaderColumn]:
    if not rows:
        return []
    row = rows[0]
    return [HeaderColumn(
            var,
            type(getattr(row, var))
            ) for var in vars(row).keys()]


def get_table_time_comparision(
    subgroups: List[KeyDriverAnalysisResultSubgroup],
    time_comparision: KeyDriverAnalysisResultTimeComparison
) -> Table:
    rows = []
    for subgroup in subgroups:
        factor_0, factor_1, factor_2 = get_factors(subgroup.factors)

        r = TimeCompareRow(
            subgroup_id=subgroup.id,
            is_top_driver=bool(subgroup.is_top_driver),
            factor_0_dimension=factor_0.dimension,
            factor_0_value=factor_0.value,
            factor_1_dimension=factor_1.dimension,
            factor_1_value=factor_1.value,
            factor_2_dimension=factor_2.dimension,
            factor_2_value=factor_2.value,
            previous_period_size=subgroup.time_comparison.previous_period_size,
            recent_period_size=subgroup.time_comparison.recent_period_size,
            previous_period_value=subgroup.time_comparison.previous_period_value,
            recent_period_value=subgroup.time_comparison.recent_period_value,
            previous_period_start=time_comparision.previous_period.start,
            previous_period_end=time_comparision.previous_period.end,
            recent_period_start=time_comparision.recent_period.start,
            recent_period_end=time_comparision.recent_period.end,
        )
        rows.append(r)

    return Table(
        build_header_from_row(rows),
        rows
    )


def get_table_group_comparision(
    subgroups: List[KeyDriverAnalysisResultSubgroup],
    group_comparision: KeyDriverAnalysisResultGroupComparison
) -> Table:
    rows = []
    for subgroup in subgroups:
        factor_0, factor_1, factor_2 = get_factors(subgroup.factors)

        r = GroupCompareRow(
            subgroup_id=subgroup.id,
            is_top_driver=bool(subgroup.is_top_driver),
            factor_0_dimension=factor_0.dimension,
            factor_0_value=factor_0.value,
            factor_1_dimension=factor_1.dimension,
            factor_1_value=factor_1.value,
            factor_2_dimension=factor_2.dimension,
            factor_2_value=factor_2.value,
            group_a_size=subgroup.group_comparison.group_a_size,
            group_b_size=subgroup.group_comparison.group_b_size,
            group_a_value=subgroup.group_comparison.group_a_value,
            group_b_value=subgroup.group_comparison.group_b_value,
            group_a_name=group_comparision.group_a.name,
            group_b_name=group_comparision.group_b.name,
        )
        rows.append(r)

    return Table(
        build_header_from_row(rows),
        rows
    )


def get_table_general_performance(
    subgroups: List[KeyDriverAnalysisResultSubgroup]
) -> Table:
    rows = []
    for subgroup in subgroups:
        factor_0, factor_1, factor_2 = get_factors(subgroup.factors)

        r = GeneralPerformanceRow(
            subgroup_id=subgroup.id,
            is_top_driver=bool(subgroup.is_top_driver),
            factor_0_dimension=factor_0.dimension,
            factor_0_value=factor_0.value,
            factor_1_dimension=factor_1.dimension,
            factor_1_value=factor_1.value,
            factor_2_dimension=factor_2.dimension,
            factor_2_value=factor_2.value,
            size=subgroup.general_performance.size,
            value=subgroup.general_performance.value,
        )
        rows.append(r)

    return Table(
        build_header_from_row(rows),
        rows
    )


def pathjoin(*args) -> str:
    return '/'.join(s.strip('/') for s in args)


def _get_table(url: str, static_analysis_id: int, auth_key: str, params: dict = {}) -> Table:
    path = ['api/v1/analyses/', str(static_analysis_id), 'runs/latest']

    url_path = build_url(url,  pathjoin(*path), params)
    
    headers = headers = {"Authorization": auth_key}

    r = requests.get(url_path, headers=headers)
    if(r.status_code != 200):
        raise Exception("Result did not complete", r)
    
    latest_analysis_response: LatestAnalysisResultResponse = LatestAnalysisResultResponse().from_dict(r.json())
    key_driver_analysis_result = latest_analysis_response.analysis_result.key_driver_analysis_result

    subgroups = key_driver_analysis_result.subgroups
    if key_driver_analysis_result.time_comparison:
        return get_table_time_comparision(subgroups, key_driver_analysis_result.time_comparison)
    elif key_driver_analysis_result.group_comparison:
        return get_table_group_comparision(subgroups, key_driver_analysis_result.group_comparison)
    elif key_driver_analysis_result.general_performance._serialized_on_wire:
        return get_table_general_performance(subgroups)
    else:
        raise ValueError("Invalid analysis_result")


def get_results(analysis_id: int, auth_key: str, params: dict = {}, auto_paginate: bool = True, url: str = 'https://vip.sisudata.com') -> Table:
    table = _get_table(url, analysis_id, auth_key, params)
    if auto_paginate:
        if not table.rows:
            return table
        if params.get('limit'):
            params['limit'] -= len(table.rows)
        if params.get('limit', 0) < 0:
            return table
        params['starting_after'] = table.rows[-1].subgroup_id

        rest_of_table = get_results(analysis_id, auth_key, params, auto_paginate, url)

        table.rows = table.rows + rest_of_table.rows
        return table
    else:
        return table



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: api.proto
# plugin: python-betterproto
from dataclasses import dataclass
from datetime import datetime
from typing import Dict, List, Optional

import betterproto


class AnalysisType(betterproto.Enum):
    """Type of Analysis eg. KEY_DRIVER."""

    ANALYSIS_TYPE_UNKNOWN = 0
    ANALYSIS_TYPE_KEY_DRIVER = 1
    ANALYSIS_TYPE_TREND = 2


class RunStatus(betterproto.Enum):
    """Status of running analysis."""

    RUN_STATUS_UNKNOWN = 0
    # Run in progress, not executing new analysis.
    RUN_STATUS_IN_FLIGHT = 1
    # Stating new analysis run execution.
    RUN_STATUS_INITIALIZING = 2
    # Analysis was stopped.
    RUN_STATUS_STOPPED = 3
    # Analysis errored out.
    RUN_STATUS_ERRORED = 4
    # Analysis completed successfully.
    RUN_STATUS_COMPLETED = 5


class AnalysisResultRunType(betterproto.Enum):
    RUN_TYPE_UNKNOWN = 0
    RUN_TYPE_SCHEDULED = 1
    RUN_TYPE_MANUAL = 2


@dataclass
class AnalysesResultRequest(betterproto.Message):
    """Request parameters for get analysis results."""

    # A limit on the number of objects to be returned, between 1 and 100. Default
    # value is 100.
    limit: Optional[int] = betterproto.message_field(1, wraps=betterproto.TYPE_UINT64)
    # starting_after is an object ID that defines your place in the list. For
    # instance, if you make a analysis list request and receive 100, ending with
    # id = 89, your subsequent call can include starting_after=89 in order to
    # fetch the next page of the list.
    starting_after: Optional[int] = betterproto.message_field(
        2, wraps=betterproto.TYPE_INT64
    )
    # Formats the return as a csv.
    csv: Optional[bool] = betterproto.message_field(3, wraps=betterproto.TYPE_BOOL)


@dataclass
class ListAnalysesRequest(betterproto.Message):
    """Request parameters for get analysis list."""

    # What type of analyses to include in the results. If not set all types will
    # be returned.
    analysis_type: "AnalysisType" = betterproto.enum_field(1)
    # A limit on the number of objects to be returned, between 1 and 100. Default
    # value is 100.
    limit: Optional[int] = betterproto.message_field(2, wraps=betterproto.TYPE_UINT64)
    # starting_after is an object ID that defines your place in the list. For
    # instance, if you make a analysis list request and receive 100, ending with
    # id = 89, your subsequent call can include starting_after=89 in order to
    # fetch the next page of the list.
    starting_after: Optional[int] = betterproto.message_field(
        3, wraps=betterproto.TYPE_INT64
    )


@dataclass
class ListAnalysesResponse(betterproto.Message):
    """ListAnalysesResponse provides list of Analyses."""

    # List of analyses.
    analyses: List["Analysis"] = betterproto.message_field(1)


@dataclass
class Analysis(betterproto.Message):
    """Provides detailed information about an analysis."""

    # Analysis id.
    id: int = betterproto.int64_field(1)
    # Analysis name.
    name: str = betterproto.string_field(2)
    # Type of Analysis eg. TYPE_KEY_DRIVER.
    type: "AnalysisType" = betterproto.enum_field(3)
    # Timestamp when the analysis was created.
    created_at: datetime = betterproto.message_field(4)


@dataclass
class RunAnalysisRequest(betterproto.Message):
    """Request payload for execute analysis workflow."""

    # Analysis id.
    id: int = betterproto.int64_field(1)


@dataclass
class RunAnalysisResponse(betterproto.Message):
    """Response message execute analysis workflow."""

    # Run id represents a new analysis run instance.
    run_id: int = betterproto.int64_field(1)
    # Indicates the current state of the analysis.
    status: "RunStatus" = betterproto.enum_field(2)


@dataclass
class LatestAnalysisResultResponse(betterproto.Message):
    """Response payload for get LatestAnalysisResult."""

    # Analysis Result.
    analysis_result: "AnalysisResult" = betterproto.message_field(1)


@dataclass
class AnalysisResult(betterproto.Message):
    """Provides details of an analysis run."""

    # The run ID Analysis run. Run ID's are unique across analyses. example:
    # 102940
    id: int = betterproto.int64_field(1)
    # Indicates if analysis run completed successfully or not.
    success: bool = betterproto.bool_field(2)
    # Time at which run was kicked off.
    requested_at: datetime = betterproto.message_field(3)
    # Time at which analysis run completed.
    completed_at: datetime = betterproto.message_field(4)
    run_type: "AnalysisResultRunType" = betterproto.enum_field(5)
    key_driver_analysis_result: "KeyDriverAnalysisResult" = betterproto.message_field(
        6, group="run_result"
    )
    trend_analysis_result: "TrendAnalysisResult" = betterproto.message_field(
        7, group="run_result"
    )


@dataclass
class KeyDriverAnalysisResult(betterproto.Message):
    """Provides details of a key driver analysis run."""

    # If subtype is TIME_COMPARISON, metadata about the time periods that are
    # compared.
    time_comparison: "KeyDriverAnalysisResultTimeComparison" = (
        betterproto.message_field(6, group="comparison")
    )
    # If subtype is GROUP_COMPARISON metadata about the groups that are being
    # compared.
    group_comparison: "KeyDriverAnalysisResultGroupComparison" = (
        betterproto.message_field(7, group="comparison")
    )
    # If the subtype is General_Performance.
    general_performance: "KeyDriverAnalysisResultGeneralPerformance" = (
        betterproto.message_field(8, group="comparison")
    )
    # Array of the subgroups selected by the key driver algorithm.
    results: List["KeyDriverAnalysisResultSubGroup"] = betterproto.message_field(9)


@dataclass
class KeyDriverAnalysisResultTimeComparison(betterproto.Message):
    """
    If subtype is TIME_COMPARISON, metadata about the time periods that are
    compared.
    """

    # The earlier of the two periods being compared.
    previous_period: "TimeRange" = betterproto.message_field(1)
    # The more recent of the two periods being compared.
    recent_period: "TimeRange" = betterproto.message_field(2)


@dataclass
class KeyDriverAnalysisResultGroupComparison(betterproto.Message):
    """
    If subtype is GROUP_COMPARISON metadata about the groups that are being
    compared.
    """

    # The first group.
    group_a: "KeyDriverAnalysisResultGroupComparisonGroupDescription" = (
        betterproto.message_field(1)
    )
    # The second group.
    group_b: "KeyDriverAnalysisResultGroupComparisonGroupDescription" = (
        betterproto.message_field(2)
    )


@dataclass
class KeyDriverAnalysisResultGroupComparisonGroupDescription(betterproto.Message):
    # The user-defined name corresponding to the first group.
    name: str = betterproto.string_field(1)


@dataclass
class KeyDriverAnalysisResultGeneralPerformance(betterproto.Message):
    """If subtype is General Performance."""

    pass


@dataclass
class KeyDriverAnalysisResultSubGroup(betterproto.Message):
    """Subgroup of a key driver analysis run."""

    # Unique ID corresponding to each subgroup, unique per analysis run.
    id: int = betterproto.int64_field(1)
    # Is false discovery rate.
    is_false_discovery_rate: Optional[bool] = betterproto.message_field(
        2, wraps=betterproto.TYPE_BOOL
    )
    # The factors that define this subgroup, represented as a map of Dimension to
    # Value.
    factors: Dict[str, "Factor"] = betterproto.map_field(
        3, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    group_comparison: "KeyDriverAnalysisResultSubGroupGroupComparisonPerformance" = (
        betterproto.message_field(4, group="details")
    )
    time_comparison: "KeyDriverAnalysisResultSubGroupTimeComparisonPerformance" = (
        betterproto.message_field(5, group="details")
    )
    general_performance: "KeyDriverAnalysisResultSubGroupGeneralPerformance" = (
        betterproto.message_field(6, group="details")
    )


@dataclass
class KeyDriverAnalysisResultSubGroupGroupComparisonPerformance(betterproto.Message):
    """
    If analysis type is GROUP_COMPARISON the metric value and size of the
    compared subgroups.
    """

    # The size of this subgroup in the first group.
    group_a_size: Optional[float] = betterproto.message_field(
        1, wraps=betterproto.TYPE_DOUBLE
    )
    # The size of this subgroup in the second group.
    group_b_size: Optional[float] = betterproto.message_field(
        2, wraps=betterproto.TYPE_DOUBLE
    )
    # The value of the metric for this of this subgroup in the first group.
    group_a_value: Optional[float] = betterproto.message_field(
        3, wraps=betterproto.TYPE_DOUBLE
    )
    # The value of the metric for this of this subgroup in the second group.
    group_b_value: Optional[float] = betterproto.message_field(
        4, wraps=betterproto.TYPE_DOUBLE
    )


@dataclass
class KeyDriverAnalysisResultSubGroupTimeComparisonPerformance(betterproto.Message):
    """
    If analysis type is TIME_COMPARISON,  the metric value and size of the
    compared subgroups.
    """

    # The size of this subgroup in the earlier of the compared periods.
    previous_period_size: Optional[float] = betterproto.message_field(
        1, wraps=betterproto.TYPE_DOUBLE
    )
    # The size of this subgroup in the more recent of the compared periods.
    recent_period_size: Optional[float] = betterproto.message_field(
        2, wraps=betterproto.TYPE_DOUBLE
    )
    # The value of the metric for this of this subgroup in the earlier of the
    # compared periods.
    previous_period_value: Optional[float] = betterproto.message_field(
        3, wraps=betterproto.TYPE_DOUBLE
    )
    # The value of the metric for this of this subgroup in the more recent of the
    # compared periods.
    recent_period_value: Optional[float] = betterproto.message_field(
        4, wraps=betterproto.TYPE_DOUBLE
    )


@dataclass
class KeyDriverAnalysisResultSubGroupGeneralPerformance(betterproto.Message):
    """
    If analysis type is GENERAL_PERFORMANCE the metric value and size for this
    subgroup.
    """

    # The size (in percent) of this subgroup relative to the overall population.
    size: Optional[float] = betterproto.message_field(1, wraps=betterproto.TYPE_DOUBLE)
    # The metric value corresponding to this subgroup.
    value: Optional[float] = betterproto.message_field(2, wraps=betterproto.TYPE_DOUBLE)


@dataclass
class TrendAnalysisResult(betterproto.Message):
    """Provides details of a Trend Analysis result."""

    # Metric level trends.
    overall_trends: List["TrendAnalysisResultTrend"] = betterproto.message_field(1)
    # Array of the subgroups in the trend.
    subgroups: List["TrendAnalysisResultSubgroup"] = betterproto.message_field(2)


@dataclass
class TrendAnalysisResultTrend(betterproto.Message):
    """Provides fields that describes the trend."""

    # Inclusive start and exclusive end time range.
    time_range: "TimeRange" = betterproto.message_field(1)
    # Y-intersept of the trend.
    intercept: float = betterproto.double_field(2)
    # Steepness of trend.
    slope: float = betterproto.double_field(3)


@dataclass
class TrendAnalysisResultSubgroup(betterproto.Message):
    """Subgroup of an trend analysis run."""

    # Unique ID corresponding to each subgroup, unique per analysis run.
    id: int = betterproto.int64_field(1)
    # The factors that define this subgroup, represented as a map of Dimension to
    # Value.
    factors: Dict[str, "Factor"] = betterproto.map_field(
        2, betterproto.TYPE_STRING, betterproto.TYPE_MESSAGE
    )
    # The size (in percent) of this subgroup relative to the overall population.
    size: Optional[float] = betterproto.message_field(3, wraps=betterproto.TYPE_DOUBLE)
    # Trends for the subgroup.
    trends: List["TrendAnalysisResultTrend"] = betterproto.message_field(4)


@dataclass
class TimeRange(betterproto.Message):
    """
    TimeRange start and end details with inclusive start and exclusive end.
    """

    start: datetime = betterproto.message_field(1)
    end: datetime = betterproto.message_field(2)


@dataclass
class Factor(betterproto.Message):
    """(Dimension, Value) pairs that define a subgroup."""

    # Value, it is either string, int, boolean, float or timestamp type.
    value: "FactorValue" = betterproto.message_field(1, group="factor_type")
    # Keyword in a text dimension.
    keyword: "FactorKeyword" = betterproto.message_field(2, group="factor_type")
    # bin of a numerical dimension.
    bin: "FactorBin" = betterproto.message_field(3, group="factor_type")


@dataclass
class FactorValue(betterproto.Message):
    """Value FactorType."""

    boolean_value: bool = betterproto.bool_field(1, group="value_type")
    integer_value: int = betterproto.int64_field(2, group="value_type")
    string_value: str = betterproto.string_field(3, group="value_type")
    float_value: float = betterproto.double_field(4, group="value_type")
    timestamp_value: datetime = betterproto.message_field(5, group="value_type")


@dataclass
class FactorKeyword(betterproto.Message):
    """Keyword FactorType."""

    keyword: str = betterproto.string_field(1)


@dataclass
class FactorBin(betterproto.Message):
    """Bin FactorType."""

    # The inclusive lower bound of the bin.
    lower_bound: Optional[float] = betterproto.message_field(
        1, wraps=betterproto.TYPE_DOUBLE
    )
    # The exclusive upper bound of the bin.
    upper_bound: Optional[float] = betterproto.message_field(
        2, wraps=betterproto.TYPE_DOUBLE
    )
    # The percentile of `lower_bound` within the factor's dimension.
    lower_bound_percentile: Optional[float] = betterproto.message_field(
        3, wraps=betterproto.TYPE_DOUBLE
    )
    # The percentile of `upper_bound` within the factor's dimension.
    upper_bound_percentile: Optional[float] = betterproto.message_field(
        4, wraps=betterproto.TYPE_DOUBLE
    )
